import asyncio
import os
import sys
from RabbitSpider.http.request import Request
from RabbitSpider.http.response import Response
from RabbitSpider.rabbit_execute import go, batch_go

if sys.platform == 'win32':
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

sys.path.append(os.path.abspath(os.path.join(os.path.abspath(sys.argv[0]), '..')))
sys.path.append(os.path.abspath(os.path.join(os.path.abspath(sys.argv[0]), '../..')))
sys.path.append(os.path.abspath(os.path.join(os.path.abspath(sys.argv[0]), '../../..')))
sys.path.append(os.path.abspath(os.path.join(os.path.abspath(sys.argv[0]), '../../../..')))

logo = r"""
    ____             __      __      _    __    _____             _        __              
   / __ \  ____ _   / /_    / /_    (_)  / /_  / ___/    ____    (_)  ____/ /  ___    _____
  / /_/ / / __ `/  / __ \  / __ \  / /  / __/  \__ \    / __ \  / /  / __  /  / _ \  / ___/
 / _, _/ / /_/ /  / /_/ / / /_/ / / /  / /_   ___/ /   / /_/ / / /  / /_/ /  /  __/ / /    
/_/ |_|  \__,_/  /_.___/ /_.___/ /_/   \__/  /____/   / .___/ /_/   \__,_/   \___/ /_/     
                                                     /_/                                   
"""
sys.stdout.write(f'\033[0;35;1m{logo}\033[0m')



from RabbitSpider.utils import event


class BasePipeline(object):
    def __init__(self, crawler):
        crawler.subscriber.subscribe(self.open_spider, event.spider_opened)
        crawler.subscriber.subscribe(self.process_item, event.spider_item)
        crawler.subscriber.subscribe(self.close_spider, event.spider_closed)

    async def open_spider(self, spider):
        """初始化数据库"""
        pass

    async def process_item(self, item, spider):
        """入库逻辑"""
        pass

    async def close_spider(self, spider):
        """关闭连接"""
        pass

    @classmethod
    def create_instance(cls, crawler):
        return cls(crawler)



import asyncio
from collections import defaultdict
from typing import Dict, Set, Callable


class Subscriber(object):
    def __init__(self):
        self._subscriber: Dict[str, Set[Callable]] = defaultdict(set)

    def subscribe(self, receiver: Callable, event: str):
        self._subscriber[event].add(receiver)

    def unsubscribe(self, receiver: Callable, event: str):
        self._subscriber[event].discard(receiver)

    def notify(self, event: str, *args, **kwargs):
        for receiver in self._subscriber[event]:
            asyncio.create_task(receiver(*args, **kwargs))

    @classmethod
    def create_instance(cls):
        return cls()



import asyncio
import sys
from typing import Type, List
from traceback import print_exc
from RabbitSpider.core.download import CurlDownload
from RabbitSpider.core.scheduler import Scheduler
from RabbitSpider.core.engine import Engine
from RabbitSpider.spider import Spider
from RabbitSpider.utils import event
from RabbitSpider.utils.subscriber import Subscriber
from RabbitSpider.utils.control import SettingManager
from asyncio.exceptions import CancelledError
from RabbitSpider.utils.control import TaskManager, PipelineManager, FilterManager


class Crawler(object):
    def __init__(self, spider: Type[Spider], mode: str, task_count: int):
        self.mode = mode
        self.task_count = task_count
        self.settings = SettingManager()
        self.subscriber = Subscriber.create_instance()
        self.spider = spider(self)
        self.scheduler = Scheduler(self.settings)
        self.filter = FilterManager(self.settings)
        self.pipelines = PipelineManager.create_instance(self)
        self.task_manager = TaskManager(self.task_count)
        self.download = CurlDownload.create_instance(self)

    async def process(self):
        self.spider.logger.info(f'任务{self.spider.name}启动')
        self.subscriber.notify(event.spider_opened, self.spider)
        try:
            async with Engine(self) as engine:
                await engine.start()
        except CancelledError as exc:
            self.spider.logger.error(f'任务{self.spider.name}中止')
            self.subscriber.notify(event.spider_error, self.spider, exc)
        except Exception as exc:
            self.spider.logger.error(f'任务{self.spider.name}中止')
            self.subscriber.notify(event.spider_error, self.spider, exc)
        else:
            self.subscriber.notify(event.spider_closed, self.spider)
            self.spider.logger.info(f'任务{self.spider.name}结束')


async def main(spider, mode, task_count):
    try:
        crawler = Crawler(spider, mode, task_count)
        await crawler.process()
    except Exception:
        print_exc()
        raise


async def go(spider: Type[Spider], mode: str = 'auto', task_count: int = 1):
    for i in sys.argv[1:]:
        key, value = i.split('=')
        if key == 'mode':
            mode = value
        if key == 'task_count':
            task_count = int(value)
    await main(spider, mode=mode, task_count=task_count)


async def batch_go(spiders: List[Type[Spider]], mode: str = 'auto', task_pool: int = 10):
    task_manager = TaskManager(task_pool)
    for spider in spiders:
        await task_manager.semaphore.acquire()
        task_manager.create_task(go(spider, mode=mode, task_count=1))
    while True:
        if task_manager.all_done():
            break
        else:
            await asyncio.sleep(1)



import pickle
import asyncio
from aio_pika import IncomingMessage
from RabbitSpider import Request
from RabbitSpider.items.item import BaseItem
from RabbitSpider.utils import event
from RabbitSpider.utils.exceptions import RabbitExpect
from collections.abc import AsyncGenerator, Coroutine, Generator
from aio_pika.exceptions import QueueEmpty, ChannelClosed, ChannelNotFoundEntity


class Engine(object):

    def __init__(self, crawler):
        self.connection = None
        self.channel = None
        self.mode = crawler.mode
        self.spider = crawler.spider
        self.subscriber = crawler.subscriber
        self.task_count = crawler.task_count
        self.scheduler = crawler.scheduler
        self.filter = crawler.filter
        self.task_manager = crawler.task_manager
        self.download = crawler.download

    async def __aenter__(self):
        self.connection, self.channel = await self.scheduler.connect()
        await self.scheduler.create_queue(self.channel, self.spider.name)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.channel.close()
        await self.connection.close()

    async def routing(self, result):
        async def rule(res):
            if isinstance(res, Request):
                if self.filter.request_seen(res):
                    self.spider.logger.info(f'生产数据：{res.to_dict()}')
                    await self.scheduler.producer(self.channel, queue=self.spider.name, body=res.to_dict())
            elif isinstance(res, BaseItem):
                self.subscriber.notify(event.spider_item, res, self.spider)

        if isinstance(result, AsyncGenerator):
            async for r in result:
                await rule(r)
        elif isinstance(result, Generator):
            for r in result:
                await rule(r)
        elif isinstance(result, Coroutine):
            r = await result
            await rule(r)
        elif isinstance(result, Request):
            await rule(result)
        elif result is None:
            pass
        else:
            raise RabbitExpect('回调函数返回类型错误！')

    async def produce(self):
        try:
            await self.scheduler.queue_purge(self.channel, self.spider.name)
            await self.routing(self.spider.start_requests())
        except Exception as e:
            self.spider.logger.error(f'{e}')
            for task in asyncio.all_tasks():
                task.cancel()

    async def crawl(self):
        while True:
            try:
                incoming_message: IncomingMessage = await self.scheduler.consumer(self.channel, self.spider.name)
            except QueueEmpty:
                if self.task_manager.all_done():
                    await self.scheduler.delete_queue(self.channel, self.spider.name)
                    break
                else:
                    continue
            except ChannelNotFoundEntity:
                break
            except ChannelClosed:
                await asyncio.sleep(1)
                self.spider.logger.warning('rabbitmq重新连接')
                self.connection, self.channel = await self.scheduler.connect()
                continue
            await self.task_manager.semaphore.acquire()
            self.task_manager.create_task(self.deal_resp(incoming_message))

    async def consume(self):
        while True:
            try:
                await self.scheduler.consumer(self.channel, queue=self.spider.name, callback=self.deal_resp,
                                              prefetch=self.task_count)
            except ChannelClosed:
                await asyncio.sleep(1)
                self.spider.logger.warning('rabbitmq重新连接')
                self.connection, self.channel = await self.scheduler.connect()
                continue
            await asyncio.Future()

    async def deal_resp(self, incoming_message: IncomingMessage):
        ret = pickle.loads(incoming_message.body)
        self.spider.logger.info(f'消费数据：{ret}')
        request, response = await self.download.send(Request(**ret))
        if response:
            try:
                callback = getattr(self.spider, ret['callback'])
                result = callback(request, response)
                if result:
                    await self.routing(result)
            except Exception as e:
                self.spider.logger.error(f'解析失败：{e}')
                for task in asyncio.all_tasks():
                    task.cancel()
            else:
                await incoming_message.ack()
        elif request:
            await self.routing(request)

    async def start(self):
        if self.mode == 'auto':
            await self.produce()
            await self.crawl()
        elif self.mode == 'm':
            await self.produce()
        elif self.mode == 'w':
            await self.consume()
        else:
            raise RabbitExpect('执行模式错误！')



from RabbitSpider import Request, Response
from RabbitSpider.utils.log import Logger
from RabbitSpider.utils import event


class Spider(object):
    name: str
    custom_settings: dict = {}

    def __init__(self, crawler):
        crawler.settings.set_dict(self.custom_settings)
        crawler.subscriber.subscribe(self.spider_opened, event.spider_opened)
        crawler.subscriber.subscribe(self.spider_closed, event.spider_closed)
        crawler.subscriber.subscribe(self.spider_error, event.spider_error)
        crawler.subscriber.subscribe(self.spider_item, event.spider_item)
        self.logger = Logger(crawler.settings, self.name)

    async def start_requests(self):
        """初始请求"""
        raise NotImplementedError

    async def parse(self, request: Request, response: Response):
        """默认回调"""
        pass

    async def spider_opened(self, *args, **kwargs):
        pass

    async def spider_closed(self, *args, **kwargs):
        pass

    async def spider_error(self, *args, **kwargs):
        pass

    async def spider_item(self, *args, **kwargs):
        pass



from curl_cffi.requests import AsyncSession
from RabbitSpider import Response, Request
from RabbitSpider.utils.control import MiddlewareManager
from RabbitSpider.utils.exceptions import RabbitExpect


class CurlDownload(object):
    def __init__(self, crawler):
        self.crawler = crawler
        self.impersonate = crawler.settings.get('IMPERSONATE')
        self.http_version = crawler.settings.get('HTTP_VERSION')
        self.middlewares = MiddlewareManager.create_instance(crawler)
        self.session = AsyncSession(verify=False)

    async def fetch(self, request):
        if request['method'].upper() == 'GET':
            res = await self.session.get(request['url'],
                                         params=request.get('params', None), cookies=request.get('cookies', None),
                                         headers=request.get('headers', None), proxy=request.get('proxy', None),
                                         allow_redirects=request.get('allow_redirects', True),
                                         http_version=self.http_version,
                                         impersonate=self.impersonate,
                                         timeout=request.get('timeout', 60)
                                         )

        elif request['method'].upper() == 'POST':
            res = await self.session.post(request['url'],
                                          data=request.get('data', None), json=request.get('json', None),
                                          cookies=request.get('cookies', None), headers=request.get('headers', None),
                                          proxy=request.get('proxy', None),
                                          http_version=self.http_version,
                                          impersonate=self.impersonate,
                                          allow_redirects=request.get('allow_redirects', True),
                                          timeout=request.get('timeout', 180))

        else:
            raise RabbitExpect(f"{request['method']}请求方式未定义，请自定义添加！")

        if res:
            response = Response(res)
            return response

    async def send(self, request: Request):
        try:
            resp = await self.middlewares.process_request(self.fetch, request)
        except Exception as exc:
            resp = await self.middlewares.process_exception(request, exc)
        if isinstance(resp, Response):
            resp = await self.middlewares.process_response(request, resp)
        if isinstance(resp, Request):
            return request, None
        if not resp:
            return None, None
        return request, resp

    @classmethod
    def create_instance(cls, crawler):
        return cls(crawler)
